#include <framerd/framerd.h>
#include <python2.2/Python.h>
#define PTR_DATA(x, y) FD_PTR_DATA(x, y) /* think this is a bug? */
#define DEFERRED_ADDRESS(ADDR) 0

FRAMERD_EXPORT int fd_revert_oid(fd_lisp oid);
FRAMERD_EXPORT int fd_commit_oid(fd_lisp oid);
typedef struct {
  PyObject o;
  fd_lisp oid;
} pyframerOID;

typedef struct {
  pyframerOID x;
} pyframerFrame;

typedef struct {
  PyObject o;
  fd_lisp vec;
} pyframerVector;

typedef struct {
  PyObject o;
  fd_lisp cons;
} pyframerCons;

typedef struct {
  PyObject o;
  fd_lisp choice;
} pyframerChoice;

typedef struct {
  PyObject o;
  fd_lisp map;
} pyframerSlotmap;

typedef struct {
  PyObject o;
  fd_pool pool;
} pyframerPool;

typedef struct {
  PyObject o;
  fd_index index;
} pyframerIndex;


staticforward PyTypeObject pyframerOID_Type;
staticforward PyTypeObject pyframerFrame_Type;
staticforward PyTypeObject pyframerVector_Type;
staticforward PyTypeObject pyframerCons_Type;
staticforward PyTypeObject pyframerChoice_Type;
staticforward PyTypeObject pyframerSlotmap_Type;
staticforward PyTypeObject pyframerPool_Type;
staticforward PyTypeObject pyframerIndex_Type;
staticforward PyObject * pyframerOID_fromFD(fd_lisp value); 
staticforward PyObject * pyframerVector_fromFD(fd_lisp value); 
staticforward PyObject * pyframerCons_fromFD(fd_lisp value); 
staticforward PyObject * pyframerChoice_fromFD(fd_lisp value); 
staticforward PyObject * pyframerSlotmap_fromFD(fd_lisp value);
staticforward PyObject * pyframerPool_fromFD(fd_pool p); 
staticforward PyObject * pyframerIndex_fromFD(fd_index p); 

fd_lisp Py2FD(PyObject *ob) {
  fd_lisp value;
  
  if (PyInt_Check(ob)) {
    value = FD_LISPFIX(PyInt_AS_LONG(ob));}
  else if (PyFloat_Check(ob)) {
    value = FD_LISPFLOAT(PyFloat_AS_DOUBLE(ob));}
  else if (PyString_Check(ob)) {
    value = fd_make_string(PyString_AsString(ob));}
  else if (ob == Py_None) {
    value = FD_VOID;}
  else if (PyObject_TypeCheck(ob, &pyframerOID_Type)) {
    value = ((pyframerOID *)ob)->oid;}
  else if (PyObject_TypeCheck(ob,  &pyframerVector_Type)) {
    value = ((pyframerVector *)ob)->vec;}
  else if (PyObject_TypeCheck(ob,  &pyframerSlotmap_Type)) {
    value = ((pyframerSlotmap *)ob)->map;}
  else if (PyObject_TypeCheck(ob,  &pyframerCons_Type)) {
    value = ((pyframerCons *)ob)->cons;}
  else if (PyObject_TypeCheck(ob,  &pyframerChoice_Type)) {
    value = ((pyframerChoice *)ob)->choice;}
  else {
    PyErr_Format(PyExc_TypeError, "No conversion method available for %s", PyString_AsString(PyObject_Repr(PyObject_Type(ob))));
  return FD_VOID;}
  return value;
}
      
    
PyObject * FD2Py(fd_lisp value) {
  
  PyObject *ob;
  switch FD_PTR_TYPE(value) {
  case fixnum_type:
    ob = PyInt_FromLong(FD_FIXLISP(value));
    break;
  case character_type:
    ob = PyString_FromStringAndSize((const char *)&FD_CHAR_CODE(value), 1);
    break;
  case symbol_type:
    ob = PyString_FromString(FD_SYMBOL_NAME(value));
    break;
  case immediate_type:
    /* maaaaaaan this sucks */
    switch (FD_PTR_DATA(value,fixnum)) {
    case FD_EMPTY_LIST_CODE: ob = pyframerCons_fromFD(value); break;
    case FD_FALSE_CODE: ob = PyInt_FromLong(0); break;
    case FD_TRUE_CODE: ob = PyInt_FromLong(1); break;
    case FD_VOID_CODE: ob = Py_None; break;
    case FD_EMPTY_CHOICE_CODE: ob = pyframerChoice_fromFD(value); break;
    case FD_LISP_EOF_CODE: ob = PyString_FromString("#EOF"); break;
    };
    break;
  case qstring_type:
  case string_type:
    ob = PyString_FromStringAndSize(FD_STRING_DATA(value), 
                                    FD_STRING_LENGTH(value));
    break;
  case flonum_type:
    ob = PyFloat_FromDouble(FD_FLOATLISP(value));
    break;
  case object_type:
    ob = pyframerOID_fromFD(value);
    break;
  case vector_type:
    ob = pyframerVector_fromFD(value);
    break;
  case pair_type:
    ob = pyframerCons_fromFD(value);
    break;
  case slotmap_type:
    ob = pyframerSlotmap_fromFD(value);
    break;  
  case choice_type:
  case proper_choice_type:
  case quoted_choice_type:
    ob = pyframerChoice_fromFD(value);
    break; 
  default:       
    PyErr_Format(PyExc_TypeError, "Unrepresentable FramerD type %d", FD_PTR_TYPE(value));
    return NULL;
  }
  Py_INCREF(ob);
  return ob;}

/** OID methods **/

static PyObject * pyframerOID_fromFD(fd_lisp value) {
  pyframerOID *self;
  if (FD_SLOTMAPP(fd_get_oid_value(value)))
    self = (pyframerOID *)PyObject_New(pyframerFrame, &pyframerFrame_Type);  
  else self = PyObject_New(pyframerOID, &pyframerOID_Type);  
  if (self == NULL) return NULL;
  fd_incref(value);
  self->oid = value;
  return (PyObject *)self;}
static PyObject * pyframerOID_new(PyObject*self, PyObject *args, PyObject *kwds) {
  char *oid_str; 
  if (!PyArg_ParseTuple(args, "s:OID", &oid_str))
    return NULL;
  return pyframerOID_fromFD(fd_parse_arg(oid_str));}
static void pyframerOID_dealloc(pyframerOID *self) {
  fd_decref(self->oid);
  PyObject_Del(self);}
static PyObject * pyframerOID_str(pyframerOID *self) {
  fd_lisp id = self->oid;     
  return PyString_FromFormat("<pyframer OID: @%x/%x>",FD_OID_ADDR_HIGH(id),FD_OID_ADDR_LOW(id));}
static PyObject * pyframerOID_repr(pyframerOID *self) {
  fd_lisp id = self->oid;     
  return PyString_FromFormat( "@%x/%x",FD_OID_ADDR_HIGH(id),FD_OID_ADDR_LOW(id));}
static PyObject * pyframerOID_get_value(pyframerOID *self) {
  return FD2Py(fd_oid_value(self->oid));
}
static PyObject * pyframerOID_set_value(pyframerOID *self, PyObject *args) {
  PyObject *ob;
  if (!PyArg_ParseTuple(args, "O:set", &ob))
     return NULL;
    fd_set_oid_value(self->oid, Py2FD(ob));
    Py_INCREF(Py_None);
  return Py_None;
}

static PyObject * pyframerOID_as_frame(pyframerOID *self) {
  pyframerFrame *f;
  if (!FD_SLOTMAPP(fd_get_oid_value(self->oid))) {
    PyErr_SetString(PyExc_TypeError, "OID does not point at a slotmap");
    return NULL;}
  f = (pyframerFrame *)PyObject_New(pyframerFrame, &pyframerFrame_Type);
  f->x.oid = self->oid;
  return (PyObject *)f;    
}
 
static PyObject * pyframerOID_commit(pyframerOID *self) {
  if (fd_commit_oid(self->oid))
    return PyInt_FromLong(1);
  else return PyInt_FromLong(0);
}

#if 0
 static PyObject * pyframerOID_revert(pyframerOID *self) {
  /* I dont think per-OID reverting works, exactly */
  if (fd_revert_oid(self->oid))
    return PyInt_FromLong(1);
  else return PyInt_FromLong(0);
} 
#endif 

static PyMethodDef pyframerOID_methods[] = {
     {"get", (PyCFunction)pyframerOID_get_value, METH_NOARGS, "Fetch the object this OID references."},
     {"set", (PyCFunction)pyframerOID_set_value, METH_VARARGS, "set(obj)\n Sets the OID's value to OBJ."},
     {"commit", (PyCFunction)pyframerOID_commit, METH_NOARGS, "Saves this OID to the pool."},
     /* {"revert", (PyCFunction)pyframerOID_revert, METH_NOARGS}, */
     {"asFrame", (PyCFunction)pyframerOID_as_frame, METH_NOARGS, "Returns an indexable Frame for this OID."},
    {NULL,      NULL} /* sentinel */};
static  PyTypeObject pyframerOID_Type = {
     PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
	0,			/*ob_size*/
	"pyframer.OID",		/*tp_name*/
	sizeof(pyframerOID),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)pyframerOID_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
        0,  /*tp_getattr*/
        0,                      /*tp_setattr*/
	0,			/*tp_compare*/
	(reprfunc)pyframerOID_repr,       /*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
        0, /*tp_as_mapping*/
	0,			/*tp_hash*/
        0,                      /*tp_call*/
        (reprfunc)pyframerOID_str,        /*tp_str*/
        0,                      /*tp_getattro*/
        0,                      /*tp_setattro*/
        0,                      /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT,     /*tp_flags*/
        0,                      /*tp_doc*/
        0,                      /*tp_traverse*/
        0,                      /*tp_clear*/
        0,                      /*tp_richcompare*/
        0,                      /*tp_weaklistoffset*/
        0,                      /*tp_iter*/
        0,                      /*tp_iternext*/
        pyframerOID_methods,                      /*tp_methods*/
        0,                      /*tp_members*/
        0,                      /*tp_getset*/
        DEFERRED_ADDRESS(&PyBaseObject_Type), /*tp_base*/
        0,                      /*tp_dict*/
        0,                      /*tp_descr_get*/
        0,                      /*tp_descr_set*/
        0,                      /*tp_dictoffset*/
        0,                      /*tp_init*/
        0,                      /*tp_alloc*/
        (newfunc)pyframerOID_new,/*tp_new*/
        0,                      /*tp_free*/
        0,                      /*tp_is_gc*/};

/** Frame methods **/

static PyObject * pyframerFrame_getitem(pyframerOID *self, PyObject *key) {
  fd_lisp value = fd_frame_get(self->oid, fd_make_symbol(PyString_AsString(key)));
  if (!FD_EMPTYP(value))
    return FD2Py(value);
  PyErr_SetObject(PyExc_KeyError, key);
  return NULL;
}

 static int pyframerFrame_setitem(pyframerOID *self, PyObject *key, PyObject *value) {
  fd_lisp k = fd_make_symbol(PyString_AsString(key));
  if (value != NULL) {
  fd_frame_set(self->oid, k, Py2FD(value));
  } else {
    PyErr_SetString(PyExc_TypeError, "object doesn't support item deletion, use .remove() instead");
      return -1;
  }  
  return 0;
}

static PyObject *pyframerFrame_add(pyframerOID *self, PyObject *args) {
  char *key;
  PyObject *value;
  if (!PyArg_ParseTuple(args, "sO:add", &key, &value))
     return NULL;
  fd_frame_add(self->oid, fd_make_symbol(key), Py2FD(value));
  Py_INCREF(Py_None);
  return Py_None;
}
static PyObject *pyframerFrame_remove(pyframerOID *self, PyObject *args) {
  char *key;
  PyObject *value;
  if (!PyArg_ParseTuple(args, "sO:remove", &key, &value))
     return NULL;
  fd_frame_remove(self->oid, fd_make_symbol(key), Py2FD(value));
  Py_INCREF(Py_None);
  return Py_None;
}
static int pyframerFrame_length(pyframerOID *self) {  
  fd_lisp m = fd_oid_value(self->oid);
  int size = FD_SLOTMAP_SIZE(m);
 return size;
}

static PyObject * pyframerFrame_keys(pyframerFrame *self) {
  PyObject *keylist = PyList_New(0);
  fd_lisp v = fd_oid_value(self->x.oid);
  FD_DO_SLOTS(slotid,value,v) {
    if (PyList_Append(keylist, FD2Py(fd_incref(slotid))) <0)
      return NULL;
  }
  return keylist;
}

static PyObject * pyframerFrame_items(pyframerFrame *self) {
  PyObject *itemlist = PyList_New(0);
  fd_lisp v = fd_oid_value(self->x.oid);
  FD_DO_SLOTS(slotid,value,v) {
    PyObject *t = PyTuple_New(2);
    PyObject *k = FD2Py(fd_incref(slotid));
    PyObject *v = FD2Py(fd_incref(value));
    Py_INCREF(k);
    Py_INCREF(v);
    PyTuple_SET_ITEM(t, 0, k);
    PyTuple_SET_ITEM(t, 1, v);       
    if (PyList_Append(itemlist, t) <0)
      return NULL;
  }
  return itemlist;
}

static PyObject * pyframerFrame_asFrame(pyframerFrame *self) {
  return (PyObject *)self;
}

static PyObject * pyframerFrame_index(pyframerFrame *self, PyObject *args) {
  pyframerIndex *index;
  int j, i = PyTuple_Size(args);
  fd_lisp slotid;
  if (!PyArg_ParseTuple(args, "O:index", &index))
    return NULL;
  if (i == 1) {
    fd_index_frame(index->index, self->x.oid);
  } else {
    for (j = 1; j <= i; j++) {
      slotid= fd_make_symbol(PyString_AsString(PyTuple_GetItem(args, j)));
      fd_index_slot_values(index->index, self->x.oid, slotid, fd_frame_get(self->x.oid, slotid));
    }    
  }
  Py_INCREF(Py_None);
  return Py_None;
}
static PyMethodDef pyframerFrame_methods[] = {
  {"asFrame", (PyCFunction)pyframerFrame_asFrame, METH_NOARGS},
  {"keys", (PyCFunction)pyframerFrame_keys, METH_NOARGS, "F.keys() -> list of F's keys"},
  {"items", (PyCFunction)pyframerFrame_items, METH_NOARGS, "F.items() -> list of F's (key, value) pairs, as 2-tuples"},
  {"add", (PyCFunction)pyframerFrame_add, METH_VARARGS, "F.add(key, val) -> None. Adds VAL to KEY's associations in F."},
  {"remove", (PyCFunction)pyframerFrame_remove, METH_VARARGS, "F.remove(key, val) -> None. Remove VAL from KEY's associations in F."},
  {"index", (PyCFunction)pyframerFrame_index, METH_VARARGS, "F.index(idx) -> None. Index F in IDX."},
  {NULL, NULL}};

static PyMappingMethods pyframerFrame_as_mapping = {
  (inquiry)pyframerFrame_length, /*mp_length*/
  (binaryfunc)pyframerFrame_getitem, /*mp_subscript*/
  (objobjargproc)pyframerFrame_setitem, /*mp_ass_subscript*/
};

static  PyTypeObject pyframerFrame_Type = {
  PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
  0,			/*ob_size*/
  "pyframer.Frame",		/*tp_name*/
  sizeof(pyframerFrame),	/*tp_basicsize*/
  0,			/*tp_itemsize*/
  /* methods */
  0, /*tp_dealloc*/
  0,			/*tp_print*/
  0,  /*tp_getattr*/
  0,                      /*tp_setattr*/
  0,			/*tp_compare*/
  0,       /*tp_repr*/
  0,			/*tp_as_number*/
  0,			/*tp_as_sequence*/
  &pyframerFrame_as_mapping, /*tp_as_mapping*/
  0,			/*tp_hash*/
  0,                      /*tp_call*/
  0,        /*tp_str*/
  0,                      /*tp_getattro*/
  0,                      /*tp_setattro*/
  0,                      /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT,     /*tp_flags*/
  0,                      /*tp_doc*/
  0,                      /*tp_traverse*/
  0,                      /*tp_clear*/
  0,                      /*tp_richcompare*/
  0,                      /*tp_weaklistoffset*/
  0,                      /*tp_iter*/
  0,                      /*tp_iternext*/
  pyframerFrame_methods, /*tp_methods*/
  0,                      /*tp_members*/
  0,                      /*tp_getset*/
  DEFERRED_ADDRESS(&pyframerOID_Type), /*tp_base*/
  0,                      /*tp_dict*/
  0,                      /*tp_descr_get*/
  0,                      /*tp_descr_set*/
  0,                      /*tp_dictoffset*/
  0,                      /*tp_init*/
  0,                      /*tp_alloc*/
  0,                       /*tp_new*/
  0,                      /*tp_free*/
  0,                      /*tp_is_gc*/
};

/** Vector methods **/

static PyObject * pyframerVector_fromFD(fd_lisp value) {
     pyframerVector *self;
     self = PyObject_New(pyframerVector, &pyframerVector_Type);
     fd_incref(value);
     self->vec = value;
     return (PyObject *)self;}
static PyObject * pyframerVector_new( PyObject*self, PyObject *args, PyObject *kwds) {
     PyObject *arg = NULL;
     int length, i;
     fd_lisp vector, item;
      if (!PyArg_ParseTuple(args, "O:Vector", &arg))
        return NULL;
      length = PySequence_Size(arg);
      vector = fd_make_vector(length);
      for (i = 0; i < length; i++) {
        item = Py2FD(PySequence_GetItem(arg, i));
        if (PyErr_Occurred()) return NULL;
        fd_incref(item);
        FD_VECTOR_SET(vector, i, item);
      }
      return pyframerVector_fromFD(vector);}
static PyObject * pyframerVector_GetItem(pyframerVector *self, int i) {
  if (i < 0 || i >= FD_VECTOR_LENGTH(self->vec)) {
    PyErr_SetString(PyExc_IndexError, "vector index out of range");
      return NULL;
      }
  return FD2Py(FD_VECTOR_REF(self->vec, i));}
int pyframerVector_SetItem(pyframerVector *self, int i, PyObject *item) {
  fd_lisp lisp_item;
  if (i < 0 || i >= FD_VECTOR_LENGTH(self->vec)) {
    PyErr_SetString(PyExc_IndexError, "vector index out of range");
    return -1;
      }
  lisp_item = Py2FD(item);
  if (PyErr_Occurred()) return -1;    
  FD_VECTOR_SET(self->vec, i, lisp_item);  
  return 0;}
static int pyframerVector_length(pyframerVector *self) {
  return FD_VECTOR_LENGTH(self->vec);}
static PyObject * pyframerVector_slice(pyframerVector *self, int ilow, int ihigh) {
  fd_lisp nl;
  int i, len;
  len = FD_VECTOR_LENGTH(self->vec);
  
  if (ilow < 0)
    ilow = 0;
  else if (ilow > len)
    ilow = len;
  if (ihigh < ilow)
    ihigh = ilow;
  else if (ihigh >len)
    ihigh = len;
  nl = fd_make_vector(ihigh - ilow);
  for (i = ilow; i < ihigh; i++) {
    fd_lisp v = FD_VECTOR_REF(self->vec, i);
    fd_incref(v);
    FD_VECTOR_SET(nl,(i - ilow), v);
  }
  return (PyObject *)pyframerVector_fromFD(nl);}
static PyObject * pyframerVector_repr(pyframerVector *self) {
  return  PyString_FromString(fd_object_to_string(self->vec));
    }
static void pyframerVector_dealloc(pyframerVector *self) {
  fd_decref(self->vec);
  PyObject_Del(self);
}

static PySequenceMethods pyframerVector_as_sequence = {
	(inquiry)pyframerVector_length,			/* sq_length */
	0,		/* sq_concat */
	0,		/* sq_repeat */
	(intargfunc)pyframerVector_GetItem,			/* sq_item */
	(intintargfunc)pyframerVector_slice,		/* sq_slice */
	(intobjargproc)pyframerVector_SetItem,		/* sq_ass_item */
        0,	/* sq_ass_slice */
        0,		/* sq_contains */
        0,	/* sq_inplace_concat */
        0,	/* sq_inplace_repeat */};

static  PyTypeObject pyframerVector_Type = {
     PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
	0,			/*ob_size*/
	"pyframer.Vector",		/*tp_name*/
	sizeof(pyframerVector),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)pyframerVector_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	0,  /*tp_getattr*/
        0,                      /*tp_setattr*/
	0,			/*tp_compare*/
	(reprfunc)pyframerVector_repr,       /*tp_repr*/
	0,			/*tp_as_number*/
	&pyframerVector_as_sequence,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
        0,                      /*tp_call*/
        0,        /*tp_str*/
        0,                      /*tp_getattro*/
        0,                      /*tp_setattro*/
        0,                      /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT,     /*tp_flags*/
        0,                      /*tp_doc*/
        0,                      /*tp_traverse*/
        0,                      /*tp_clear*/
        0,                      /*tp_richcompare*/
        0,                      /*tp_weaklistoffset*/
        0,                      /*tp_iter*/
        0,                      /*tp_iternext*/
        0,                      /*tp_methods*/
        0,                      /*tp_members*/
        0,                      /*tp_getset*/
        DEFERRED_ADDRESS(&PyBaseObject_Type), /*tp_base*/
        0,                      /*tp_dict*/
        0,                      /*tp_descr_get*/
        0,                      /*tp_descr_set*/
        0,                      /*tp_dictoffset*/
        0,                      /*tp_init*/
        0,                      /*tp_alloc*/
        (newfunc)pyframerVector_new, /*tp_new*/
        0,                      /*tp_free*/
        0,                      /*tp_is_gc*/};

/** Cons methods **/
static PyObject * pyframerCons_fromFD(fd_lisp value) {
  pyframerCons *self;
  self = PyObject_New(pyframerCons, &pyframerCons_Type);
  fd_incref(value);
  self->cons = value;
  return (PyObject *)self;}

static PyObject * pyframerCons_new(PyObject*self, PyObject *args, PyObject *kwds) {
  PyObject *car, *cdr;
  fd_lisp cons;
  if (!PyArg_ParseTuple(args, "OO:Cons", &car, &cdr)) {
    return NULL;}
  cons = FD_MAKE_PAIR(fd_incref(Py2FD(car)), fd_incref(Py2FD(cdr)));
  return pyframerCons_fromFD(cons);}

static void pyframerCons_setcar(pyframerCons *self, PyObject *args) {
  PyObject *newcar;
  fd_lisp car;
  if (PyArg_ParseTuple(args, "O:setcar", &newcar)) {
    fd_decref(FD_CAR(self->cons));
    car = Py2FD(newcar);
    fd_incref(car);
    FD_RPLACA(self->cons, car);}}

static void pyframerCons_setcdr(pyframerCons *self, PyObject *args) {
  PyObject *newcdr;
  fd_lisp cdr;
  if ( PyArg_ParseTuple(args, "O:setcdr", &newcdr)) {
    fd_decref(FD_CDR(self->cons));
    cdr = Py2FD(newcdr);
    fd_incref(cdr);
    FD_RPLACD(self->cons, cdr);}}

static PyObject * pyframerCons_car(pyframerCons *self) {
  if (!FD_PRIM_TYPEP(self->cons,pair_type))
    return (PyObject *)self;
  else return FD2Py(FD_CAR(self->cons));
}  
static PyObject * pyframerCons_cdr(pyframerCons *self) {
  if (!FD_PRIM_TYPEP(self->cons,pair_type))
    return (PyObject *)self;
  return FD2Py(FD_CDR(self->cons));}
static void pyframerCons_dealloc(pyframerCons *self) {
  fd_decref(self->cons);
  PyObject_Del(self);
}

static PyObject * pyframerCons_repr(pyframerCons *self) {
  return  PyString_FromString(fd_object_to_string(self->cons));
}

static PyMethodDef pyframerCons_methods[] = {
  {"setcar", (PyCFunction)pyframerCons_setcar, METH_VARARGS, "C.setcar(val) -> None. Sets C's car to VAL."},
  {"setcdr", (PyCFunction)pyframerCons_setcdr, METH_VARARGS,  "C.setcdr(val) -> None. Sets C's cdr to VAL."},
  {"car", (PyCFunction)pyframerCons_car, METH_NOARGS, "C.car() -> first value in C"},
  {"cdr", (PyCFunction)pyframerCons_cdr, METH_NOARGS, "C.cdr() -> second value in C"},
  {NULL, NULL}};

static PyTypeObject pyframerCons_Type = {
     PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
	0,			/*ob_size*/
	"pyframer.cons",		/*tp_name*/
	sizeof(pyframerCons),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)pyframerCons_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	0,                      /*tp_getattr*/
        0,                      /*tp_setattr*/
	0,			/*tp_compare*/
        (reprfunc)pyframerCons_repr,       /*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
        0,                      /*tp_call*/
        0,        /*tp_str*/
        0,                      /*tp_getattro*/
        0,                      /*tp_setattro*/
        0,                      /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT,     /*tp_flags*/
        0,                      /*tp_doc*/
        0,                      /*tp_traverse*/
        0,                      /*tp_clear*/
        0,                      /*tp_richcompare*/
        0,                      /*tp_weaklistoffset*/
        0,                      /*tp_iter*/
        0,                      /*tp_iternext*/
        pyframerCons_methods, /*tp_methods*/
        0,                      /*tp_members*/
        0,                      /*tp_getset*/
        DEFERRED_ADDRESS(&PyBaseObject_Type), /*tp_base*/
        0,                      /*tp_dict*/
        0,                      /*tp_descr_get*/
        0,                      /*tp_descr_set*/
        0,                      /*tp_dictoffset*/
        0,                      /*tp_init*/
        0,                      /*tp_alloc*/
        (newfunc)pyframerCons_new, /*tp_new*/
        0,                      /*tp_free*/
        0,                      /*tp_is_gc*/
};

static PyObject * pyframerChoice_fromFD(fd_lisp value) {
  pyframerChoice *self;
  self = PyObject_New(pyframerChoice, &pyframerChoice_Type);
  fd_incref(value);
  self->choice = value;
  return (PyObject *)self;}

static PyObject * pyframerChoice_new(PyObject*self, PyObject *args, PyObject *kwds) {
  PyObject *ob;
  fd_lisp choice = FD_EMPTY_CHOICE;
  int i, j;
  if (!PyArg_ParseTuple(args, "O:Choice", &ob)) {
    return NULL;}
  i = PySequence_Length(ob);
  for (j=0; j < i; j++) {
    FD_ADD_TO_CHOICE(choice, fd_incref(Py2FD(PySequence_GetItem(ob, j))));
  }
  return pyframerChoice_fromFD(fd_return_proper_choice(choice));
}

static PyObject * pyframerChoice_repr(pyframerChoice *self) {
  return  PyString_FromString(fd_object_to_string(self->choice));
}

static void pyframerChoice_dealloc(pyframerChoice *self) {
  fd_decref(self->choice);
  PyObject_Del(self);
}
static PyObject * pyframerChoice_iter(pyframerChoice *self) {
  PyObject * tup = PyTuple_New(FD_CHOICE_SIZE(self->choice));
  PyObject * it, *ob;
  int i=0;
  FD_DO_CHOICES(x, self->choice) {
    ob =  FD2Py(x);
    PyTuple_SetItem(tup, i, ob);
    i++;
  } FD_END_DO_CHOICES;      
  it = PySeqIter_New(tup);
  return (PyObject *)it;
}
static int pyframerChoice_length(pyframerChoice *self) {
  return FD_CHOICE_SIZE(self->choice);}

static PySequenceMethods pyframerChoice_as_sequence = {
  (inquiry)pyframerChoice_length, 0, 0, 0, 0, 0, 0, 0, 0, 0};
static PyTypeObject pyframerChoice_Type = {
     PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
	0,			/*ob_size*/
	"pyframer.Choice",		/*tp_name*/
	sizeof(pyframerChoice),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)pyframerChoice_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	0,                      /*tp_getattr*/
        0,                      /*tp_setattr*/
	0,			/*tp_compare*/
        (reprfunc)pyframerChoice_repr,       /*tp_repr*/
	0,			/*tp_as_number*/
	&pyframerChoice_as_sequence, /*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
        0,                      /*tp_call*/
        0,        /*tp_str*/
        0,                      /*tp_getattro*/
        0,                      /*tp_setattro*/
        0,                      /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT,     /*tp_flags*/
        0,                      /*tp_doc*/
        0,                      /*tp_traverse*/
        0,                      /*tp_clear*/
        0,                      /*tp_richcompare*/
        0,                      /*tp_weaklistoffset*/
        (getiterfunc)pyframerChoice_iter, /*tp_iter*/
        0,                      /*tp_iternext*/
        0, /*tp_methods*/
        0,                      /*tp_members*/
        0,                      /*tp_getset*/
        DEFERRED_ADDRESS(&PyBaseObject_Type), /*tp_base*/
        0,                      /*tp_dict*/
        0,                      /*tp_descr_get*/
        0,                      /*tp_descr_set*/
        0,                      /*tp_dictoffset*/
        0,                      /*tp_init*/
        0,                      /*tp_alloc*/
        (newfunc)pyframerChoice_new, /*tp_new*/
        0,                      /*tp_free*/
        0,                      /*tp_is_gc*/};


/** Slotmap methods **/

static PyObject * pyframerSlotmap_fromFD(fd_lisp value) {
  pyframerSlotmap *self;
  self = PyObject_New(pyframerSlotmap, &pyframerSlotmap_Type);
  fd_incref(value);
  self->map = value;
  return (PyObject *)self;}

static PyObject * pyframerSlotmap_getitem(pyframerSlotmap *self, PyObject *key) {
  fd_lisp value = fd_frame_get(self->map, fd_make_symbol(PyString_AsString(key)));
  if (!FD_EMPTYP(value))
    return FD2Py(value);
  PyErr_SetObject(PyExc_KeyError, key);
  return NULL;
}

static int pyframerSlotmap_setitem(pyframerSlotmap *self, PyObject *key, PyObject *value) {
  fd_frame_set(self->map, fd_make_symbol(PyString_AsString(key)), Py2FD(value));
    return 0;
    }

static PyObject * pyframerSlotmap_new(PyObject *self, PyObject *args, PyObject *kwds) {
  PyObject *k, *v;
  int pos = 0;
  if (!PyArg_ParseTuple(args, ":Slotmap")) return NULL;
  self = pyframerSlotmap_fromFD(fd_make_slotmap(8));
  if (!kwds) return self;
  while (PyDict_Next(kwds, &pos, &k, &v)) {
    pyframerSlotmap_setitem((pyframerSlotmap *)self, k, v);}
  return self;
}
static int  pyframerSlotmap_length(pyframerSlotmap *self) {
  fd_lisp m = self->map;
  int size = FD_SLOTMAP_SIZE(m);
  return size;
}

static void pyframerSlotmap_dealloc(pyframerSlotmap *self) {
  fd_decref(self->map);
  PyObject_Del(self);
}

static PyMappingMethods pyframerSlotmap_as_mapping = {
	(inquiry)pyframerSlotmap_length , /*mp_length*/
	(binaryfunc)pyframerSlotmap_getitem, /*mp_subscript*/
	(objobjargproc)pyframerSlotmap_setitem, /*mp_ass_subscript*/
};

static PyObject * pyframerSlotmap_repr(pyframerSlotmap *self) {
  return  PyString_FromString(fd_object_to_string(self->map));
}

static PyTypeObject pyframerSlotmap_Type = {
  PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
  0,			/*ob_size*/
  "pyframer.Slotmap",		/*tp_name*/
  sizeof(pyframerSlotmap),	/*tp_basicsize*/
  0,			/*tp_itemsize*/
  /* methods */
  (destructor)pyframerSlotmap_dealloc, /*tp_dealloc*/
  0,			/*tp_print*/
  0,                      /*tp_getattr*/
  0,                      /*tp_setattr*/
  0,			/*tp_compare*/
  (reprfunc)pyframerSlotmap_repr,       /*tp_repr*/
  0,			/*tp_as_number*/
  0,			/*tp_as_sequence*/
  &pyframerSlotmap_as_mapping, /*tp_as_mapping*/
  0,			/*tp_hash*/
  0,                      /*tp_call*/
  0,        /*tp_str*/
  0,                      /*tp_getattro*/
  0,                      /*tp_setattro*/
  0,                      /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT,     /*tp_flags*/
  0,                      /*tp_doc*/
  0,                      /*tp_traverse*/
  0,                      /*tp_clear*/
  0,                      /*tp_richcompare*/
  0,                      /*tp_weaklistoffset*/
  0,                      /*tp_iter*/
  0,                      /*tp_iternext*/
  0, /*tp_methods*/
  0,                      /*tp_members*/
  0,                      /*tp_getset*/
  DEFERRED_ADDRESS(&PyBaseObject_Type), /*tp_base*/
  0,                      /*tp_dict*/
  0,                      /*tp_descr_get*/
  0,                      /*tp_descr_set*/
  0,                      /*tp_dictoffset*/
  0,                      /*tp_init*/
  0,                      /*tp_alloc*/
  (newfunc)pyframerSlotmap_new, /*tp_new*/
  0,                      /*tp_free*/
  0,                      /*tp_is_gc*/
};

/** Pool methods **/

static PyObject * pyframerPool_fromFD(fd_pool p) {
  pyframerPool *self;
  self = PyObject_New(pyframerPool, &pyframerPool_Type);
  self->pool = p;
  return (PyObject *)self;}
static PyObject * pyframerPool_new(pyframerPool *type, PyObject *args, PyObject *kwds) {
  char * pool_name;
  pyframerPool * self;
  if (!PyArg_ParseTuple(args, "s:Pool", &pool_name))
    return NULL;
    self = PyObject_New(pyframerPool, &pyframerPool_Type);
    FD_WITH_HANDLING {
      self->pool = fd_use_pool(pool_name);
    } FD_ON_EXCEPTION {
      PyErr_SetString(PyExc_Exception, fd_theException());
      return NULL;
     FD_END_HANDLING};
    return (PyObject *)self;}
static PyObject *pyframerPool_commit(pyframerPool *self) {
  fd_commit_pool(self->pool);
  Py_INCREF(Py_None);
  return Py_None;}
static PyObject *pyframerPool_revert(pyframerPool *self) {
  fd_revert_pool(self->pool);
    Py_INCREF(Py_None);
    return Py_None;}
static PyObject *  pyframerPool_load(pyframerPool *self) {
  return PyInt_FromLong(fd_pool_load(self->pool));}
static PyObject * pyframerPool_base(pyframerPool *self) {
  return pyframerOID_fromFD(fd_make_oid(self->pool->base));}
static PyObject * pyframerPool_capacity(pyframerPool *self) {
  return PyInt_FromLong(self->pool->capacity);}
static PyObject * pyframerPool_nextOID(pyframerPool *self) {
  return pyframerOID_fromFD(fd_new_oid(self->pool));}
static PyObject * pyframerPool_nextFrame(pyframerPool *self) {
  pyframerFrame *f;
  f = (pyframerFrame *)PyObject_New(pyframerFrame, &pyframerFrame_Type);
  f->x.oid = fd_new_oid(self->pool);
  fd_set_oid_value(f->x.oid, fd_make_slotmap(8));
  return (PyObject *)f;    
}
static void pyframerPool_dealloc(pyframerPool *self) {
  fd_commit_pool(self->pool);
  fd_close_pool(self->pool);
  PyObject_Del(self);}

static PyMethodDef pyframerPool_methods[] = {
  {"load", (PyCFunction)pyframerPool_load, METH_NOARGS, "P.load() -> The number of OIDs that have been allocated in P."},
  {"base", (PyCFunction)pyframerPool_base, METH_NOARGS, "P.base() -> The first OID in P."},
  {"commit", (PyCFunction)pyframerPool_commit, METH_NOARGS, "P.commit() -> None. Save all modified OIDs."},
  {"revert", (PyCFunction)pyframerPool_revert, METH_NOARGS, "P.revert() -> None. Discard modifications to this pool since the last commit."},
  {"capacity", (PyCFunction)pyframerPool_capacity, METH_NOARGS, "P.capacity() -> The maximum number of OIDs allowed in P."},
  {"nextOID", (PyCFunction)pyframerPool_nextOID, METH_NOARGS, "P.nextOID() -> Allocates a fresh OID and returns it."},
{"nextFrame", (PyCFunction)pyframerPool_nextFrame, METH_NOARGS, "P.nextFrame() -> Allocates a fresh OID, associate it with a Slotmap, and returns it as a Frame."},
  {NULL,      NULL} /* sentinel */};

static PyTypeObject pyframerPool_Type = {
     PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
	0,			/*ob_size*/
	"pyframer.Pool",		/*tp_name*/
	sizeof(pyframerPool),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)pyframerPool_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	0,                      /*tp_getattr*/
        0,                      /*tp_setattr*/
	0,			/*tp_compare*/
        0,       /*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
        0, /*tp_as_mapping*/
	0,			/*tp_hash*/
        0,                      /*tp_call*/
        0,        /*tp_str*/
        0,                      /*tp_getattro*/
        0,                      /*tp_setattro*/
        0,                      /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT,     /*tp_flags*/
        0,                      /*tp_doc*/
        0,                      /*tp_traverse*/
        0,                      /*tp_clear*/
        0,                      /*tp_richcompare*/
        0,                      /*tp_weaklistoffset*/
        0,                      /*tp_iter*/
        0,                      /*tp_iternext*/
        pyframerPool_methods, /*tp_methods*/
        0,                      /*tp_members*/
        0,                      /*tp_getset*/
        DEFERRED_ADDRESS(&PyBaseObject_Type), /*tp_base*/
        0,                      /*tp_dict*/
        0,                      /*tp_descr_get*/
        0,                      /*tp_descr_set*/
        0,                      /*tp_dictoffset*/
        0,                      /*tp_init*/
        0,                      /*tp_alloc*/
        (newfunc)pyframerPool_new, /*tp_new*/
        0,                      /*tp_free*/
        0,                      /*tp_is_gc*/
};

/** Index methods **/


static PyObject * pyframerIndex_fromFD(fd_index p) {
  pyframerIndex *self;
  self = PyObject_New(pyframerIndex, &pyframerIndex_Type);
  self->index = p;
  return (PyObject *)self;}
static PyObject * pyframerIndex_new(pyframerIndex *type, PyObject *args, PyObject *kwds) {
  char * index_name;
  pyframerIndex * self;
  if (!PyArg_ParseTuple(args, "s:Index", &index_name))
    return NULL;
  self = PyObject_New(pyframerIndex, &pyframerIndex_Type);
  FD_WITH_HANDLING {
    self->index = fd_open_index(index_name);
  } FD_ON_EXCEPTION {
    PyErr_SetString(PyExc_Exception, fd_theException());
    return NULL;
    FD_END_HANDLING};
  return (PyObject *)self;}

static PyObject *pyframerIndex_get(pyframerIndex *self, PyObject *args) {
  PyObject *k;
  PyObject *dflt = FD2Py(FD_EMPTY_CHOICE);
    if (!PyArg_ParseTuple(args, "O|O:get", &k, &dflt))
      return NULL;    
    return pyframerChoice_fromFD(fd_index_get(self->index, fd_make_symbol(PyString_AsString(k)), Py2FD(dflt)));
    
}

static PyObject *pyframerIndex_zap(pyframerIndex *self, PyObject *args) {
  PyObject *k;
  PyObject *dflt = FD2Py(FD_EMPTY_CHOICE);
    if (!PyArg_ParseTuple(args, "O|O:zap", &k, &dflt))
      return NULL;
    fd_index_drop(self->index,fd_make_symbol(PyString_AsString(k)),fd_index_get(self->index,fd_make_symbol(PyString_AsString(k)),FD_EMPTY_CHOICE));
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *pyframerIndex_set(pyframerIndex *self, PyObject *args) {
  PyObject *k, *v;
     if (!PyArg_ParseTuple(args, "O|O:set", &k, &v))
      return NULL;
     fd_index_set(self->index, fd_make_symbol(PyString_AsString(k)), Py2FD(v));
     Py_INCREF(Py_None);
     return Py_None;
}

static PyObject *pyframerIndex_add(pyframerIndex *self, PyObject *args) {
  PyObject *k, *v;
     if (!PyArg_ParseTuple(args, "O|O:add", &k, &v))
      return NULL;
    fd_index_add(self->index,fd_make_symbol(PyString_AsString(k)) , Py2FD(v));
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject * pyframerIndex_keys(pyframerIndex *self) {
  return FD2Py(fd_index_keys(self->index));
}

static PyObject * pyframerIndex_find(pyframerIndex *self, PyObject *args, PyObject *kwds) {
  PyObject *k, *v;
  fd_lisp l;
  int pos = 0;

  l = FD_EMPTY_LIST;
  if (!kwds || (PyMapping_Length(kwds) == 0))
    Py_INCREF(Py_None);
    return Py_None;
  while (PyDict_Next(kwds, &pos, &k, &v)) {
    l = FD_MAKE_PAIR(fd_make_symbol(PyString_AsString(k)), FD_MAKE_PAIR(Py2FD(v),l));
  }
  return pyframerChoice_fromFD(fd_strict_search(fd_make_cptr(index_type,self->index), l));

}

static PyObject * pyframerIndex_findSimilar(pyframerIndex *self, PyObject *args) {
  pyframerFrame *f;
  pyframerChoice *slots;
  if (!PyArg_ParseTuple(args, "O|O:findSimilar", &f, &slots))
    return NULL;
  return pyframerChoice_fromFD(fd_find_similar(fd_make_cptr(index_type, self->index), f->x.oid, slots ? slots->choice : FD_EMPTY_CHOICE));

}
static PyObject *pyframerIndex_commit(pyframerIndex *self) {
  fd_commit_index(self->index);
  Py_INCREF(Py_None);
  return Py_None;}
static PyObject *pyframerIndex_revert(pyframerIndex *self) {
  fd_revert_index(self->index);
  Py_INCREF(Py_None);
  return Py_None;}
static void pyframerIndex_dealloc(pyframerIndex *self) {
  fd_commit_index(self->index);
  fd_close_index(self->index);
  PyObject_Del(self);}

static PyMethodDef pyframerIndex_methods[] = {
  {"get", (PyCFunction)pyframerIndex_get, METH_VARARGS},
  {"add", (PyCFunction)pyframerIndex_add, METH_VARARGS},
  {"set", (PyCFunction)pyframerIndex_set, METH_VARARGS},
  {"zap", (PyCFunction)pyframerIndex_zap, METH_VARARGS},
  {"keys", (PyCFunction)pyframerIndex_keys, METH_NOARGS, "I.keys() -> All keys in I."},
  {"commit", (PyCFunction)pyframerIndex_commit, METH_NOARGS, "I.commit() -> None. Saves the index to disk."},
  {"revert", (PyCFunction)pyframerIndex_revert, METH_NOARGS, "I.revert() -> None. Discards uncommitted changes to I."},
  {"find", (PyCFunction)pyframerIndex_find, METH_KEYWORDS, "I.find(k1=v1, k2=v2, ...) -> all indexed frames with requested key-value pairs." },
  {"findSimilar", (PyCFunction)pyframerIndex_findSimilar, METH_VARARGS, "I.findSimilar(frame) -> the indexed frames that have the most key-value pairs in common with FRAME."},
  {NULL,      NULL} /* sentinel */};

static PyTypeObject pyframerIndex_Type = {
     PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
	0,			/*ob_size*/
	"pyframer.Index",		/*tp_name*/
	sizeof(pyframerIndex),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)pyframerIndex_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	0,                      /*tp_getattr*/
        0,                      /*tp_setattr*/
	0,			/*tp_compare*/
        0,       /*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
        0, /*tp_as_mapping*/
	0,			/*tp_hash*/
        0,                      /*tp_call*/
        0,        /*tp_str*/
        0,                      /*tp_getattro*/
        0,                      /*tp_setattro*/
        0,                      /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT,     /*tp_flags*/
        0,                      /*tp_doc*/
        0,                      /*tp_traverse*/
        0,                      /*tp_clear*/
        0,                      /*tp_richcompare*/
        0,                      /*tp_weaklistoffset*/
        0,                      /*tp_iter*/
        0,                      /*tp_iternext*/
        pyframerIndex_methods, /*tp_methods*/
        0,                      /*tp_members*/
        0,                      /*tp_getset*/
        DEFERRED_ADDRESS(&PyBaseObject_Type), /*tp_base*/
        0,                      /*tp_dict*/
        0,                      /*tp_descr_get*/
        0,                      /*tp_descr_set*/
        0,                      /*tp_dictoffset*/
        0,                      /*tp_init*/
        0,                      /*tp_alloc*/
        (newfunc)pyframerIndex_new, /*tp_new*/
        0,                      /*tp_free*/
        0,                      /*tp_is_gc*/
};

static PyObject * pyframerList(PyObject *self, PyObject *args) {
  fd_lisp head, new, old, current;
  int i=0, len=0;
  len = PyTuple_Size(args);  
  if (len == 0)
    return pyframerCons_fromFD(FD_EMPTY_LIST);
  head = current = FD_MAKE_PAIR(FD_EMPTY_LIST, FD_EMPTY_LIST);
  while( i < len) {
    new = FD_MAKE_PAIR(FD_EMPTY_LIST, FD_EMPTY_LIST);  
    FD_RPLACA(current, fd_incref(Py2FD(PyTuple_GET_ITEM(args, i))));
    FD_RPLACD(current, fd_incref(new));
    old = current;
    current = new;
    i=i+1;
  }
  fd_decref(FD_CDR(old));
  FD_RPLACD(old, FD_EMPTY_LIST);
  return pyframerCons_fromFD(head);
}

static PyObject * pyframerCreatePool(PyObject *self, PyObject *args) {
  int size = 1;
  int minsize = 16000;
  char * filename = NULL;
  char * label = NULL;
  fd_lisp sm = fd_make_slotmap(8);
  fd_pool p;
  if (!PyArg_ParseTuple(args, "s|is", &filename, &minsize, &label))
    return NULL;

  while (size < (1<<28))
    if (size >= minsize) break; 
    else size=size*2;

  fd_make_file_pool(filename, 
                    fd_allocate_pool(FD_STRING_DATA(fd_getenv("SUPER_POOL")),
                                     size,filename),
                    size, 1, FD_EMPTY_CHOICE);
  
  if (label) 
    fd_label_file_pool(filename, fd_make_string(label));
  
  p = fd_use_pool(filename);
  fd_set_oid_value(fd_make_oid(p->base), sm);
  fd_frame_set(sm, fd_make_symbol("FREELIST"), FD_EMPTY_LIST);
  fd_frame_set(sm, fd_make_symbol("PYFRAMER"), FD_TRUE);
  return pyframerPool_fromFD(p);
}
    
static PyObject *nil = NULL;
static PyMethodDef pyframer_methods[] = {
  {"List", (PyCFunction)pyframerList, METH_VARARGS, "pyframer.List(*items) -> a linked list composed of ITEMS."},
  {"createPool", (PyCFunction)pyframerCreatePool, METH_VARARGS, "pyframer.createPool(filename) -> a newly created pool, stored in the named file."},
  {NULL, NULL}};

static char module_doc[] = "";

DL_EXPORT(void) initpyframer(void) {

  PyObject *m, *d; 
  PyType_Ready(&pyframerOID_Type);
  PyType_Ready(&pyframerVector_Type);
  PyType_Ready(&pyframerPool_Type);
  PyType_Ready(&pyframerCons_Type);
  PyType_Ready(&pyframerChoice_Type);
  pyframerFrame_Type.tp_base = &pyframerOID_Type;
  PyType_Ready(&pyframerFrame_Type);
  PyType_Ready(&pyframerSlotmap_Type);
  PyType_Ready(&pyframerIndex_Type);
  m = Py_InitModule3("pyframer", pyframer_methods, module_doc);
  d = PyModule_GetDict(m);
  PyDict_SetItemString(d, "OID", (PyObject *)&pyframerOID_Type);
  PyDict_SetItemString(d, "Frame", (PyObject *)&pyframerFrame_Type);
  PyDict_SetItemString(d, "Vector", (PyObject *)&pyframerVector_Type);
  PyDict_SetItemString(d, "Pool", (PyObject *)&pyframerPool_Type);
  PyDict_SetItemString(d, "Index", (PyObject *)&pyframerIndex_Type);
  PyDict_SetItemString(d, "Cons", (PyObject *)&pyframerCons_Type);
  nil = pyframerCons_fromFD(FD_EMPTY_LIST);
  PyDict_SetItemString(d, "nil", (PyObject *)nil);
  PyDict_SetItemString(d, "Choice", (PyObject *)&pyframerChoice_Type);
  PyDict_SetItemString(d, "Slotmap", (PyObject *)&pyframerSlotmap_Type);
  fd_inhibit_herald(1);
  setenv("FD_DIE_CONVENTIONALLY","1",1);
  fd_set_session_mnemonic("pyframer");
  fd_initialize_framerd();
}
